﻿Imports System.Drawing
Imports System.Drawing.Printing
Imports System.Collections.Generic
Imports System.Windows.Forms
Imports System.Text
Imports System.Runtime.InteropServices
Public Class RtfClass
#Region "fields"
    Private MyRtb As RichTextBox
    Private MyFirstChar As Integer
    Private CurrentPage As Integer
    Private PageCount As Integer
    Private Rc As Rectangle
    Private NextY As Integer
    Private RtfPrinted As Boolean
#End Region

#Region "Properties"
    Public WriteOnly Property SetRtf As RichTextBox
        Set(value As RichTextBox)
            MyRtb = value
        End Set
    End Property
    Public WriteOnly Property SetRtfLocation As Rectangle
        Set(value As Rectangle)
            Rc = value
        End Set
    End Property
    Public ReadOnly Property EndOfText As Boolean
        Get
            Return RtfPrinted
        End Get
    End Property
    Public ReadOnly Property NextPosition As Integer
        Get
            Return NextY
        End Get
    End Property
#End Region
    Public Sub PrintRtf(e As PrintPageEventArgs)
        RtfPrinted = False
        If PageCount < 0 Then
            PageCount = GetPageCount(e)
        End If
        CurrentPage += 1
        ' render text
        Dim fmt As FORMATRANGE = GetFormatRange(e, MyFirstChar)
        Dim nextChar As Integer = FormatMyRange(MyRtb, True, fmt)
        e.Graphics.ReleaseHdc(fmt.hdc)

        'Get bottom of rectangle to set next print position for non-rtf data on the same page
        NextY = (fmt.rc.bottom - fmt.rc.top + Rc.Y) / 14.4
        
        If nextChar >= MyRtb.TextLength Then
            RtfPrinted = True
        End If
    End Sub

#Region "implementation"
    ' build a FORMATRANGE structure with the proper page size and hdc

    Private Function GetFormatRange(e As PrintPageEventArgs, firstChar As Integer) As FORMATRANGE
        ' get supplied rectangle in twips
        Rc.X = CInt(Rc.X * 14.4 + 0.5)
        Rc.Y = CInt(Rc.Y * 14.4 + 0.5)
        Rc.Width = CInt(Rc.Width * 14.4 + 0.5)
        Rc.Height = CInt(Rc.Height * 14.4 + 0.5)

        ' set up FORMATRANGE structure
        Dim hdc = e.Graphics.GetHdc()
        Dim fmt = New FORMATRANGE()
        fmt.hdc = InlineAssignHelper(fmt.hdcTarget, hdc)
        fmt.rc.SetRect(Rc)
        fmt.rcPage = fmt.rc
        fmt.cpMin = firstChar
        fmt.cpMax = -1

        Return fmt
    End Function

    ' get a page count by using FormatRange to measure the content
    Private Function GetPageCount(e As PrintPageEventArgs) As Integer
        Dim pageCount As Integer = 0

        Dim fmt As FORMATRANGE = GetFormatRange(e, 0)
        Dim firstChar As Integer = 0
        While firstChar < MyRtb.TextLength
            fmt.cpMin = firstChar
            firstChar = FormatMyRange(MyRtb, False, fmt)
            pageCount += 1
        End While
        e.Graphics.ReleaseHdc(fmt.hdc)

        Return pageCount
    End Function

#End Region

#Region "Win32 API"

    ' messages used by RichEd20.dll
    Const WM_USER As Integer = &H400, EM_FORMATRANGE As Integer = WM_USER + &H39

    ' Win32 RECT
    <StructLayout(LayoutKind.Sequential)> _
    Private Structure RECT
        Public left As Integer, top As Integer, right As Integer, bottom As Integer
        Public Sub SetRect(rc As Rectangle)
            left = rc.Left
            top = rc.Top
            right = rc.Right
            bottom = rc.Bottom
        End Sub
    End Structure

    ' FORMATRANGE is used by RichEd20.dll to render RTF
    <StructLayout(LayoutKind.Sequential)> _
    Private Structure FORMATRANGE
        Public hdc As IntPtr, hdcTarget As IntPtr
        Public rc As RECT, rcPage As RECT
        Public cpMin As Integer, cpMax As Integer
    End Structure

    ' send the EM_FORMATRANGE message to the RichTextBox to render or measure
    ' a range of the document into a target specified by a FORMATRANGE structure.
    Private Function FormatMyRange(rtb As RichTextBox, render As Boolean, ByRef fmt As FORMATRANGE) As Integer
        ' render
        Dim nextChar As Integer = SendMessageFormatRange(rtb.Handle, EM_FORMATRANGE, If(render, 1, 0), fmt)
        ' Find the real size of the last portion of RTF text
        While nextChar > MyRtb.TextLength
            Rc.Height -= 28.8
            fmt.rc.SetRect(Rc)
            nextChar = SendMessageFormatRange(rtb.Handle, EM_FORMATRANGE, If(render, 1, 0), fmt)
        End While
        ' reset
        SendMessage(rtb.Handle, EM_FORMATRANGE, 0, 0)

        ' return next character to print
        Return nextChar
    End Function

    ' two flavors of SendMessage
    <DllImport("USER32.DLL", CharSet:=CharSet.Auto)> _
    Private Shared Function SendMessage(hWnd As IntPtr, wMsg As UInteger, wParam As Integer, lParam As Integer) As Integer
    End Function

    <DllImport("USER32.DLL", CharSet:=CharSet.Auto, EntryPoint:="SendMessage")> _
    Private Shared Function SendMessageFormatRange(hWnd As IntPtr, wMsg As UInteger, wParam As Integer, ByRef lParam As FORMATRANGE) As Integer
    End Function
    Private Shared Function InlineAssignHelper(Of T)(ByRef target As T, value As T) As T
        target = value
        Return value
    End Function
#End Region
End Class
